import qrotor as qr
import numpy as np


def test_phase():
    sys = qr.System()
    sys.B = 1.0
    sys.potential_name = 'cos'
    sys.gridsize = 10000
    sys.solve()
    # plus pi/2, which will be -3pi/2
    sys.change_phase(0.5)
    assert round(sys.grid[0], 2) == round(-np.pi * 3/2, 2)
    # The first potential value should be 0,
    # but remember that the potential offset is corrected
    # so it should be half potential_max, so 1.0/2
    assert round(sys.potential_values[0], 2) == 0.5
    # minus pi, which will become -pi/2
    sys.change_phase(-1)
    assert round(sys.grid[0], 2) == round(-np.pi/2, 2)
    assert round(sys.potential_values[0], 2) == 0.5
    # Were eigenvalues calculated?
    assert len(sys.eigenvalues) > 0




"""
.. include:: ../_README_temp.md
"""


from ._version import __version__ as version
from .system import System
from .constants import *
from . import systems
from . import rotation
from . import potential
from . import solve
from . import plot




import qrotor as qr
import aton


folder = 'tests/samples/'
structure = folder + 'CH3NH3.in'
structure_120 = folder + 'CH3NH3_120.in'
structure_60 = folder + 'CH3NH3_60.in'


def test_save_and_load():
    system = qr.System()
    system.gridsize = 36
    system.potential_name = 'sin'
    system.B = 1
    system.solve_potential()
    potential_file = folder + '_temp_potential.csv'
    # Remove the file if it exists
    try:
        aton.file.remove(potential_file)
    except:
        pass
    qr.potential.save(system, comment='hi', filepath=potential_file)
    system_new = qr.potential.load(potential_file)
    assert system_new.gridsize == system.gridsize
    assert round(system_new.potential_values[0], 5) == round(system.potential_values[0], 5)
    assert round(system_new.potential_values[5], 5) == round(system.potential_values[5], 5)
    assert round(system_new.potential_values[13], 5) == round(system.potential_values[13], 5)
    assert system_new.comment == 'hi'
    aton.file.remove(potential_file)
    # If we don't provide a comment, it should be the name of the folder
    system.comment = None
    qr.potential.save(system, filepath=potential_file)
    system_new = qr.potential.load(potential_file)
    assert system_new.comment == 'samples'
    aton.file.remove(potential_file)




"""
# Description

The `System` object contains all the information needed for a single QRotor calculation.
This class can be loaded directly as `qrotor.System()`.

---
"""


import numpy as np
from .constants import *
from aton import alias
from ._version import __version__


class System:
    """Quantum system.

    Contains all the data for a single QRotor calculation, with both inputs and outputs.

    Energy units are in meV and angles are in radians, unless stated otherwise.
    """
    def __init__(
            self,
            comment: str = None,
            B: float = B_CH3,
            gridsize: int = 200000,
            searched_E: int = 21,
            correct_potential_offset: bool = True,
            save_eigenvectors: bool = True,
            potential_name: str = '',
            potential_constants: list = None,
            tags: str = '',
            ):
        """A new quantum system can be instantiated as `system = qrotor.System()`.
        This new system will contain the default values listed above.
        """
        ## Technical
        self.version = __version__
        """Version of the package used to generate the data."""
        self.comment: str = comment
        """Custom comment for the dataset."""
        self.searched_E: int = searched_E
        """Number of energy eigenvalues to be searched."""
        self.correct_potential_offset: bool = correct_potential_offset
        """Correct the potential offset as `V - min(V)` or not."""
        self.save_eigenvectors: bool = save_eigenvectors
        """Save or not the eigenvectors. Final file size will be bigger."""
        self.tags: str = tags
        """Custom tags separated by spaces, such as the molecular group, etc.

        Can be used to filter between datasets.
        """
        ## Potential
        self.B: float = B
        """Kinetic rotational energy, as in $B=\\frac{\\hbar^2}{2I}$.

        Defaults to the value for a methyl group.
        """
        self.gridsize: int = gridsize
        """Number of points in the grid."""
        self.grid = []
        """The grid with the points to be used in the calculation.

        Can be set automatically over $2 \\pi$ with `System.set_grid()`.
        Units must be in radians.
        """
        self.potential_name: str = potential_name
        """Name of the desired potential: `'zero'`, `'titov2023'`, `'test'`...

        If empty or unrecognised, the custom potential values inside `System.potential_values` will be used. 
        """
        self.potential_constants: list = potential_constants
        """List of constants to be used in the calculation of the potential energy, in the `qrotor.potential` module."""
        self.potential_values = []
        """Numpy [ndarray](https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html) with the potential values for each point in the grid.

        Can be calculated with a function available in the `qrotor.potential` module,
        or loaded externally with the `qrotor.potential.load()` function.
        Potential energy units must be in meV.
        """
        # Potential values determined upon solving
        self.potential_offset: float = None
        """`min(V)` before offset correction when `correct_potential_offset = True`"""
        self.potential_min: float = None
        """`min(V)`"""
        self.potential_max: float = None
        """`max(V)`"""
        # Energies determined upon solving
        self.eigenvectors = []
        """Eigenvectors, if `save_eigenvectors` is True. Beware of the file size."""
        self.eigenvalues = []
        """Calculated eigenvalues of the system. In meV."""
        self.E_levels: list = []
        """List of `eigenvalues` grouped by energy levels, found below `potential_max`."""
        self.deg: float = None
        """Estimated degeneracy of the `E_levels` found below `potential_max`."""
        self.E_activation: float = None
        """Activation energy or energy barrier, from the ground torsional state to the top of the potential barrier, `max(V) - min(eigenvalues)`"""
        self.excitations: list = []
        """Torsional excitations, as the difference between each energy level with respect to the ground state.

        Considers the means between degenerated eigenvalues for all energy levels below `potential_max`.
        """
        self.splittings: list = []
        """Tunnel splitting energies, for every degenerated energy level.
        
        Calculated for all `E_levels` as the difference between
        the mean of the eigenvalues from A and the mean of the eigenvalues from E,
        see [R. M. Dimeo, American Journal of Physics 71, 885–893 (2003)](https://doi.org/10.1119/1.1538575).
        """
        self.runtime: float = None
        """Time taken to solve the eigenvalues."""

    def solve(self, gridsize:int=None, B:int=None):
        """Default user method to solve the quantum system.

        The potential can be interpolated to a new `gridsize`.

        Same as running `qrotor.solve.energies(System)`
        with an optional new gridsize.
        """
        from .solve import energies
        if gridsize:
            self.gridsize = gridsize
        if B:
            self.B = B
        return energies(self)

    def solve_potential(self, gridsize:int=None):
        """Default user method to quickly solve the potential of the quantum system.

        This method does not solve the energies of the system,
        it just computes the potential and sets `System.potential_max`,
        `System.potential_min` and `System.potential_offset` accordingly.
        To solve the potential AND the energies, check `System.solve()`.

        The potential can be interpolated to a new `gridsize`.

        Same as running `qrotor.solve.potential(System)`
        with an optional new gridsize.
        """
        from .solve import potential
        if gridsize:
            self.gridsize = gridsize
        return potential(self)

    def change_phase(self, phase:float, calculate:bool=True):
        """Apply a phase shift to the grid and potential values.

        The `phase` should be a multiple of $\\pi$ (e.g., 3/2 for $3\\pi/2$).
        The resulting grid will be expressed between $-2\\pi$ and $2\\pi$.

        The System is solved immediately after the phase change.
        This last step ensures that all eigenvalues and wavefunctions are correct.
        You can override this step with `calculate = False`,
        but remember to solve the System later!
        """
        if not any(self.potential_values) or not any(self.grid):
            raise ValueError("System.potential_values and System.grid must be set before applying a phase shift.")
        # Normalise the phase between 0 and 2
        if abs(phase) >= 2:
            phase = phase % 2
        while phase < 0:
            phase = phase + 2
        # Shift the grid, between -2pi and 2pi
        self.grid = (self.grid + (phase * np.pi))
        # Apply the phase shift to potential values
        phase_points = int((phase / 2) * self.gridsize)
        self.potential_values = np.roll(self.potential_values, phase_points)
        # Check that the grid is still within -2pi and 2pi, otherwise normalise it for a final time
        while self.grid[0] <= (-2 * np.pi + 0.1):  # With a small tolerance
            self.grid = self.grid + 2 * np.pi
        while self.grid[-1] >= 2.5 * np.pi:  # It was not a problem until reaching 5/2 pi
            self.grid = self.grid -2 * np.pi
        print(f'Potential shifted by {phase}π')
        if calculate:
            self.solve()
        return self

    def set_grid(self, gridsize:int=None):
        """Sets the `System.grid` to the specified `gridsize` from 0 to $2\\pi$.

        If the system had a previous grid and potential values,
        it will interpolate those values to the new gridsize,
        using `qrotor.potential.interpolate()`.
        """
        if gridsize == self.gridsize:
            return self  # Nothing to do here
        if gridsize:
            self.gridsize = gridsize
        # Should we interpolate?
        if any(self.potential_values) and any(self.grid) and self.gridsize:
            from .potential import interpolate
            self = interpolate(self)
        # Should we create the values from zero?
        elif self.gridsize:
                self.grid = np.linspace(0, 2*np.pi, self.gridsize)
        else:
            raise ValueError('gridsize must be provided if there is no System.gridsize')
        return self

    def reduce_size(self):
        """Discard data that takes too much space,
        like eigenvectors, potential values and grids."""
        self.eigenvectors = []
        self.potential_values = []
        self.grid = []
        return self

    def summary(self):
        """Returns a dict with a summary of the System data."""
        return {
            'version': self.version,
            'comment': self.comment,
            'tags': self.tags,
            'searched_E': self.searched_E,
            'correct_potential_offset': self.correct_potential_offset,
            'save_eigenvectors': self.save_eigenvectors,
            'B': self.B,
            'gridsize': self.gridsize,
            'potential_name': self.potential_name,
            'potential_constants': self.potential_constants.tolist() if isinstance(self.potential_constants, np.ndarray) else self.potential_constants,
            'potential_offset': self.potential_offset,
            'potential_min': self.potential_min,
            'potential_max': self.potential_max,
            'eigenvalues': self.eigenvalues.tolist() if isinstance(self.eigenvalues, np.ndarray) else self.eigenvalues,
            'E_levels': self.E_levels,
            'deg': self.deg,
            'excitations': self.excitations,
            'splittings': self.splittings,
            'E_activation': self.E_activation,
            'runtime': self.runtime,
        }




"""
# Description

This module provides straightforward functions to plot QRotor data.


# Index

| | |
| --- | --- |
| `potential()`        | Potential values as a function of the angle |
| `energies()`         | Calculated eigenvalues |
| `reduced_energies()` | Reduced energies E/B as a function of the reduced potential V/B |
| `wavefunction()`     | Selected wavefunctions or squared wavefunctions of a system |
| `splittings()`       | Tunnel splitting energies of a list of systems |
| `convergence()`      | Energy convergence of a list of systems calculated with different parameters |

---
"""


from .system import System
from . import systems
from . import constants
import matplotlib.pyplot as plt
import numpy as np
from copy import deepcopy
import aton.alias as alias


def potential(
        data:System|list,
        title:str=None,
        marker:str|list='',
        linestyle:str|list='-',
        cm:str=None,
        cm_range:tuple|list=(0.0,0.8),
        color:list|tuple|str=None,
        normalize:bool=False,
        ylim:tuple=None,
        rc:dict={},
        ) -> None:
    """Plot the potential values of `data` (System object, or list of systems).

    Title can be customized with `title`.
    If empty, system[0].comment will be used as title if no more comments are present.

    `marker` and `linestyle` can be a Matplotlib string or list of strings.
    Custom colormaps can be set through `cm`, in a custom `cm_range` from 0 to 1.
    Custom colors can also be set through `color`.
    Alternatively, custom color lists can be set through the `color` variable.

    Set `normalize = True` to normalize by their respective `qrotor.system.System.potential_max`.
    This can be useful if you have performed subtractions or similar operations.
    In this case, you might also want to play with `ylim` to adjust the y-axis limits.

    Additional matplotlib runtime configuration
    [rcParams](https://matplotlib.org/stable/api/matplotlib_configuration_api.html#matplotlib.RcParams)
    can be set with the `rc` dict.
    """
    data_copy = deepcopy(data)
    system = systems.as_list(data_copy)
    title_str = title if title else (system[0].comment if (system[0].comment and (len(system) == 1 or not system[-1].comment)) else 'Rotational potential energy')
    # Marker as a list
    if isinstance(marker, list):
        if len(marker) < len(system):
            marker.extend([''] * (len(system) - len(marker)))
    else:
        marker = [marker] * len(system)
    # Linestyle as a list
    if isinstance(linestyle, list):
        if len(linestyle) < len(system):
            linestyle.extend(['-'] * (len(system) - len(linestyle)))
    else:
        linestyle = [linestyle] * len(system)
    # Custom colormap or colors
    colors = None
    if isinstance(cm, str):
        cmap = plt.get_cmap(cm)
        colors = cmap(np.linspace(cm_range[0], cm_range[1], len(system)))
    elif color:
        colors = color
        if isinstance(colors, list):
            if len(colors) < len(system):
                colors.extend(['grey'] * (len(system) - len(colors)))
        else:
            colors = [colors] * len(system)

    with plt.rc_context(rc):
        plt.figure(layout='constrained')
        plt.title(title_str)
        plt.xlabel('Angle / rad')
        plt.ylabel('Potential energy / meV')
        if normalize:
            plt.ylabel('Energy / V$_{3}$')
            for s in system:
                s.potential_values = s.potential_values / s.potential_max
        if ylim:
            plt.ylim(ylim)

        plt.xticks([-2*np.pi, -3*np.pi/2, -np.pi, -np.pi/2, 0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi], [r'$-2\pi$', r'$-\frac{3\pi}{2}$', r'$-\pi$', r'$-\frac{\pi}{2}$', '0', r'$\frac{\pi}{2}$', r'$\pi$', r'$\frac{3\pi}{2}$', r'$2\pi$'])

        if colors is not None:
            for i, s in enumerate(system):
                plt.plot(s.grid, s.potential_values, marker=marker[i], linestyle=linestyle[i], label=s.comment, color=colors[i])
        else:  # Regular plot
            for i, s in enumerate(system):
                plt.plot(s.grid, s.potential_values, marker=marker[i], linestyle=linestyle[i], label=s.comment)

        if all(s.comment for s in system) and len(system) != 1:
            leg = plt.legend()
            leg.set_draggable(True)

        plt.show()


def energies(
        data,
        title:str=None,
        rc:dict={},
        ) -> None:
    """Plot the eigenvalues of `data` (System or a list of System objects).

    You can use up to 1 tag per system to differentiate between molecular groups.

    Additional matplotlib runtime configuration
    [rcParams](https://matplotlib.org/stable/api/matplotlib_configuration_api.html#matplotlib.RcParams)
    can be set with the `rc` dict.
    """
    if isinstance(data, System):
        var = [data]
    else:  # Should be a list
        systems.as_list(data)
        var = data

    V_colors = ['C0', 'C1', 'C2', 'C3', 'C4']
    E_colors = ['lightblue', 'sandybrown', 'lightgrey', 'lightcoral', 'plum']
    E_linestyles = ['--', ':', '-.']
    edgecolors = E_colors

    V_linestyle = '-'
    title = title if title else (var[0].comment if var[0].comment else 'Energy eigenvalues')
    ylabel_text = f'Energy / meV'
    xlabel_text = 'Angle / radians'

    with plt.rc_context(rc):
        plt.figure(layout='constrained')
        plt.xlabel(xlabel_text)
        plt.ylabel(ylabel_text)
        plt.title(title)
        plt.xticks([-2*np.pi, -3*np.pi/2, -np.pi, -np.pi/2, 0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi], [r'$-2\pi$', r'$-\frac{3\pi}{2}$', r'$-\pi$', r'$-\frac{\pi}{2}$', '0', r'$\frac{\pi}{2}$', r'$\pi$', r'$\frac{3\pi}{2}$', r'$2\pi$'])

        unique_potentials = []
        unique_groups = []
        for i, system in enumerate(var):
            V_color = V_colors[i % len(V_colors)]
            E_color = E_colors[i % len(E_colors)]
            E_linestyle = E_linestyles[i % len(E_linestyles)]
            edgecolor = edgecolors[i % len(edgecolors)]

            # Plot potential energy if it is unique
            if not any(np.array_equal(system.potential_values, value) for value in unique_potentials):
                unique_potentials.append(system.potential_values)
                plt.plot(system.grid, system.potential_values, color=V_color, linestyle=V_linestyle)

            # Plot eigenvalues
            if any(system.eigenvalues):
                text_offset = 3 * len(unique_groups)
                if system.tags not in unique_groups:
                    unique_groups.append(system.tags)
                for j, energy in enumerate(system.eigenvalues):
                    plt.axhline(y=energy, color=E_color, linestyle=E_linestyle)
                    # Textbox positions are a bit weird when plotting more than 2 systems, but whatever...
                    plt.text(j%3*1.0 + text_offset, energy, f'$E_{{{j}}}$ = {round(energy,4):.04f}', va='top', bbox=dict(edgecolor=edgecolor, boxstyle='round,pad=0.2', facecolor='white', alpha=0.8))
                if len(systems.get_tags(var)) > 1:
                    plt.plot([], [], color=E_color, label=f'{system.tags} Energies')  # Add to legend

        if len(systems.get_tags(var)) > 1:
            plt.subplots_adjust(right=0.85)
            leg = plt.legend() #(bbox_to_anchor=(1.1, 0.5), loc='center', fontsize='small')
            leg.set_draggable(True)

        plt.show()


def reduced_energies(
        data:list,
        title:str=None,
        values:list=[],
        legend:list=[],
        rc:dict={},
        ) -> None:
    """Plots the reduced energy of the system E/B vs the reduced potential energy V/B.

    Takes a `data` list of System objects as input.
    An optional `title` can be specified.

    Optional maximum reduced potential `values` are plotted
    as vertical lines (floats or ints) or regions
    (lists inside the values list, from min to max).
    A `legend` of the same len as `values` can be included.
    These values are assumed to be divided by B by the user.

    Additional matplotlib runtime configuration
    [rcParams](https://matplotlib.org/stable/api/matplotlib_configuration_api.html#matplotlib.RcParams)
    can be set with the `rc` dict.
    """
    if values and (isinstance(values, float) or isinstance(values, int) or isinstance(values, np.float64)):
        values = [values]
    if values and len(values) <= len(legend):
        plot_legend = True
    else:
        plot_legend = False
        legend = [''] * len(values)
    systems.as_list(data)
    title = title if title else (data[0].comment if data[0].comment else 'Reduced energies')
    number_of_levels = data[0].searched_E
    x = []
    for system in data:
        potential_max_B = system.potential_max / system.B
        x.append(potential_max_B)
    colors = plt.cm.viridis(np.linspace(0, 1, number_of_levels+1))  # +1 to avoid the lighter tones

    with plt.rc_context(rc):
        plt.figure(layout='constrained')

        for i in range(number_of_levels):
            y = []
            for system in data:
                eigenvalues_B_i = system.eigenvalues[i] / system.B
                y.append(eigenvalues_B_i)
            plt.plot(x, y, marker='', linestyle='-', color=colors[i])
        # Add vertical lines in the specified values
        line_colors = plt.cm.tab10(np.linspace(0, 1, len(values)))
        for i, value in enumerate(values):
            if isinstance(value, list):
                min_value = min(value)
                max_value = max(value)
                plt.axvspan(min_value, max_value, color=line_colors[i], alpha=0.2, linestyle='', label=legend[i])
            else:
                plt.axvline(x=value, color=line_colors[i], linestyle='--', label=legend[i], alpha=0.5)
        plt.xlabel('V$_{B}$ / B')
        plt.ylabel('E / B')
        plt.title(title)
        if plot_legend:
            leg = plt.legend()
            leg.set_draggable(True)
        plt.show()


def wavefunction(
        system:System,
        title:str=None,
        square:bool=True,
        levels=[0, 1, 2],
        overlap=False,
        yticks:bool=False,
        rc:dict={},
        ) -> None:
    """Plot the wavefunction of a `system` for the specified `levels`.

    Wavefunctions are squared by default, showing the probabilities;
    To show the actual wavefunctions, set `square = False`.

    `levels` can be a list of indexes, or the number of levels to plot.

    Specific wavefunctions can be overlapped with `overlap` as a list with the target indexes.
    The `overlap` value can also be the max number of wavefunctions to add.
    All found wavefunctions can be added together with `overlap = True`;
    but note that this overlap is limited by the number of System.searched_E,
    that must be specified before solving the system.
    Setting `overlap` will ignore the `levels` argument.

    Set `yticks = True` to plot the wavefunction yticks.

    Additional matplotlib runtime configuration
    [rcParams](https://matplotlib.org/stable/api/matplotlib_configuration_api.html#matplotlib.RcParams)
    can be set with the `rc` dict.
    """
    data = deepcopy(system)
    eigenvectors = data.eigenvectors
    title = title if title else (data.comment if data.comment else 'System wavefunction')
    with plt.rc_context(rc):
        fig, ax1 = plt.subplots(layout='constrained')
        plt.title(title)
        ax1.set_xlabel('Angle / radians')
        ax1.set_ylabel('Potential / meV')
        ax1.set_xticks([-2*np.pi, -3*np.pi/2, -np.pi, -np.pi/2, 0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi], [r'$-2\pi$', r'$-\frac{3\pi}{2}$', r'$-\pi$', r'$-\frac{\pi}{2}$', '0', r'$\frac{\pi}{2}$', r'$\pi$', r'$\frac{3\pi}{2}$', r'$2\pi$'])
        ax1.plot(data.grid, data.potential_values, color='blue', linestyle='-')
        ax2 = ax1.twinx()
        if not yticks:
            ax2.set_yticks([])
        ax2.set_ylabel('Squared wavefunction' if square else 'Wavefunction')
        # Set levels list
        if isinstance(levels, int) or isinstance(levels, float):
            levels = [x for x in range(int(levels))]
        if not isinstance(levels, list):
            raise ValueError('levels must be an int or a list of ints')
        # Set overlap if requested
        if overlap == True and isinstance(overlap, bool):
            eigenvectors = [np.sum(eigenvectors, axis=0)]
            levels = [0]
            show_legend = False
        elif overlap is not False and (isinstance(overlap, int) or isinstance(overlap, float)):
            max_int = int(overlap)
            eigenvectors = [np.sum(eigenvectors[:max_int], axis=0)]
            levels = [0]
            show_legend = False
        elif isinstance(overlap, list):
            eigenvectors = [np.sum([eigenvectors[i] for i in overlap], axis=0)]
            levels = [0]
            show_legend = False
        else:
            show_legend = True
        # Square values if so
        if square:
            eigenvectors = [vec**2 for vec in eigenvectors]
        # Plot the wavefunction
        for i in levels:
            ax2.plot(data.grid, eigenvectors[i], linestyle='--', label=f'{i}')
        if show_legend:
            leg = fig.legend()#(loc='upper right', bbox_to_anchor=(0.9, 0.88), fontsize='small', title='Index')
            leg.set_draggable(True)
        plt.show()


def splittings(
        data:list,
        title:str=None,
        units:str='ueV',
        rc:dict={},
        ) -> None:
    """Plot the tunnel splitting energies of a `data` list of systems.

    The different `System.comment` are shown in the horizontal axis.
    An optional `title` can be specified.
    Default units shown are $\\mu$eV (`'ueV'`).
    Available units are: `'ueV'`, `'meV'`, `'Ry'`, or `'B'` (free rotor units).

    Additional matplotlib runtime configuration
    [rcParams](https://matplotlib.org/stable/api/matplotlib_configuration_api.html#matplotlib.RcParams)
    can be set with the `rc` dict.
    """
    title = title if title != None else 'Tunnel splitting energies'
    calcs = deepcopy(data)
    calcs = systems.as_list(calcs)

    with plt.rc_context(rc):
        fig, ax = plt.subplots(layout='constrained')
        ax.set_ylabel("Energy / meV")

        y = [c.splittings[0] for c in calcs]
        x = [c.comment for c in calcs]
        # What units do we want?
        if units.lower() in alias.units['ueV']:
            y = [j * 1000 for j in y]  # Convert meV to micro eV
            ax.set_ylabel("Energy / $\\mu$eV")
        elif units.lower() in alias.units['Ry']:
            y = [j * constants.meV_to_Ry for j in y]
            ax.set_ylabel("Energy / Ry")
        elif units.upper() == 'B':
            y = [j / c.B for j, c in zip(y, calcs)]
            ax.set_ylabel("Energy / B")
        #else:  # It's okay let's use meV

        ax.bar(range(len(y)), y)
        for i, comment in enumerate(x):
            ax.text(x=i, y=0, s=comment+' ', rotation=45, verticalalignment='top', horizontalalignment='right')
        ax.set_xlabel("")
        ax.set_title(title)
        ax.set_xticks([])
        fig.tight_layout()
        plt.show()


def convergence(data:list, rc:dict={}) -> None:
    """Plot the energy convergence of a `data` list of Systems as a function of the gridsize.

    Additional matplotlib runtime configuration
    [rcParams](https://matplotlib.org/stable/api/matplotlib_configuration_api.html#matplotlib.RcParams)
    can be set with the `rc` dict.
    """
    systems.as_list(data)
    gridsizes = [system.gridsize for system in data]
    runtimes = [system.runtime for system in data]
    deviations = []  # List of lists, containing all eigenvalue deviations for every system
    searched_E = data[0].searched_E
    for system in data:
        deviation_list = []
        for i, eigenvalue in enumerate(system.eigenvalues):
            ideal_E = systems.calculate_ideal_E(i)
            deviation = abs(ideal_E - eigenvalue)
            deviation_list.append(deviation)
        deviation_list = deviation_list[1:]  # Remove ground state
        deviations.append(deviation_list)
    # Plotting
    with plt.rc_context(rc):
        fig, ax1 = plt.subplots(layout='constrained')
        ax1.set_xlabel('Grid size')
        ax1.set_ylabel('Error / meV')
        ax1.set_xscale('log')
        ax1.set_yscale('log')
        ax2 = ax1.twinx()
        ax2.set_ylabel('Runtime / s')
        ax2.set_yscale('log')
        ax2.plot(gridsizes, runtimes, color='tab:grey', label='Runtime', linestyle='--')
        colors = plt.cm.viridis(np.linspace(0, 1, searched_E))  # Should be searched_E-1 but we want to avoid lighter colors
        for i in range(searched_E-1):
            if i % 2 == 0:  # Ignore even numbers, since those levels are degenerated.
                continue
            ax1.plot(gridsizes, [dev[i] for dev in deviations], label=f'$E_{{{int((i+1)/2)}}}$', color=colors[i])
        leg = fig.legend()#(loc='upper right', bbox_to_anchor=(0.9, 0.88), fontsize='small')
        leg.set_draggable(True)
        plt.title(data[0].comment if data[0].comment else 'Energy convergence vs grid size')
        plt.show()




import qrotor as qr
import aton.api as api
import aton.txt.extract as extract
import aton.file as file


folder = 'tests/samples/'
structure = folder + 'CH3NH3.in'
structure_120 = folder + 'CH3NH3_120.in'
structure_60 = folder + 'CH3NH3_60.in'


def test_rotation():
    CH3 = [
        '0.100   0.183   0.316',
        '0.151   0.532   0.842',
        '0.118   0.816   0.277',
    ]
    # 120 degrees (it should remain the same)
    qr.rotation.rotate_qe(filepath=structure, positions=CH3, angle=120, precision=2)
    for coord in CH3:
        rotated_coord = api.pwx.get_atom(filepath=structure_120, position=coord, precision=2)
        rotated_coord = extract.coords(rotated_coord)
        coord = extract.coords(coord)
        rotated_coord_rounded = []
        coord_rounded = []
        for i in rotated_coord:
            rotated_coord_rounded.append(round(i, 2))
        for i in coord:
            coord_rounded.append(round(i, 2))
        assert coord_rounded == rotated_coord_rounded
    file.remove(structure_120)

    # 60 degrees (it should change quite a lot)
    ideal = [
        '0.146468644022416   0.837865866372631   0.641449758215011',
        '0.095062781582172   0.488975944606740   0.115053787468686',
        '0.128156574395412   0.205890189020629   0.680672454316303',
    ]
    qr.rotation.rotate_qe(filepath=structure, positions=CH3, angle=60, precision=2)
    for coord in ideal:
        rotated_coord = api.pwx.get_atom(filepath=structure_60, position=coord, precision=3)
        rotated_coord = extract.coords(rotated_coord)
        coord = extract.coords(coord)
        rotated_coord_rounded = []
        coord_rounded = []
        for i in rotated_coord:
            rotated_coord_rounded.append(round(i, 2))
        for i in coord:
            coord_rounded.append(round(i, 2))
        assert coord_rounded == rotated_coord_rounded
    file.remove(structure_60)




"""
# Description

Common constants and default inertia values used in QRotor.

Bond lengths and angles were obtained from MAPbI3, see
[K. Drużbicki *et al*., Crystal Growth & Design 24, 391–404 (2024)](https://doi.org/10.1021/acs.cgd.3c01112).
"""


import numpy as np
import periodictable
import scipy.constants as const


# Quick conversion factors
Ry_to_eV = const.physical_constants['Rydberg constant times hc in eV'][0]
"""Quick conversion factor from Rydberg to eV energy."""
Ry_to_meV = Ry_to_eV * 1000
"""Quick conversion factor from Rydberg to meV energy."""
eV_to_Ry = 1 / Ry_to_eV
"""Quick conversion factor from eV to Rydberg."""
meV_to_Ry = 1 / Ry_to_meV
"""Quick conversion factor from meV to Rydberg."""
cm1_to_meV = (const.h * const.c * 100 / const.e) * 1000
"""Quick conversion factor from cm$^{-1}$ to meV."""
meV_to_cm1 = 1/cm1_to_meV
"""Quick conversion factor from meV to cm$^{-1}$."""
amu_to_kg = const.physical_constants['atomic mass constant'][0]
"""Quick conversion factor from amu to kg."""
kg_to_amu = 1 / amu_to_kg
"""Quick conversion factor from kg to amu."""

# Distance between Carbon and Hydrogen atoms (measured from MAPbI3)
distance_CH = 1.09285   # Angstroms
"""Distance of the C-H bond, in Angstroms."""
distance_NH = 1.040263  # Angstroms
"""Distance of the N-H bond, in Angstroms."""

# Angles between atoms:  C-C-H  or  N-C-H  etc (from MAPbI3)
angle_CH_external = 108.7223
"""External angle of the X-C-H bond, in degrees."""
angle_NH_external = 111.29016
"""External angle of the X-N-H bond, in degrees."""
angle_CH = 180 - angle_CH_external
"""Internal angle of the X-C-H bond, in degrees."""
angle_NH = 180 - angle_NH_external
"""Internal angle of the X-N-H bond, in degrees."""

# Rotation radius (calculated from distance and angle)
r_CH = distance_CH * np.sin(np.deg2rad(angle_CH)) * 1e-10
"""Rotation radius of the methyl group, in meters."""
r_NH = distance_NH * np.sin(np.deg2rad(angle_NH)) * 1e-10
"""Rotation radius of the amine group, in meters."""

# Inertias, SI units
I_CH3 = 3 * (periodictable.H.mass * amu_to_kg * r_CH**2)
"""Inertia of CH3, in kg·m^2."""
I_CD3 = 3 * (periodictable.D.mass * amu_to_kg * r_CH**2)
"""Inertia of CD3, in kg·m^2."""
I_NH3 = 3 * (periodictable.H.mass * amu_to_kg * r_NH**2)
"""Inertia of NH3, in kg·m^2."""
I_ND3 = 3 * (periodictable.D.mass * amu_to_kg * r_NH**2)
"""Inertia of ND3, in kg·m^2."""

# Inertias of the co-rotation, SI units
I_CH3NH3 = I_CH3 + I_NH3
"""Inertia of CH3NH3+, in kg·m^2."""
I_CD3NH3 = I_CD3 + I_NH3
"""Inertia of CD3NH3+, in kg·m^2."""
I_CH3ND3 = I_CH3 + I_ND3
"""Inertia of CH3ND3+, in kg·m^2."""
I_CD3ND3 = I_CD3 + I_ND3
"""Inertia of CD3ND3+, in kg·m^2."""

# Inertias of the dis-rotation, SI units
I_CH3NH3_dis = 1 / ((1/I_CH3) + (1/I_NH3))
"""Inertia of the disrotatory torsion of CH3NH3+, in kg·m^2."""
I_CD3NH3_dis = 1 / ((1/I_CD3) + (1/I_NH3))
"""Inertia of the disrotatory torsion of CD3NH3+, in kg·m^2."""
I_CH3ND3_dis = 1 / ((1/I_CH3) + (1/I_ND3))
"""Inertia of the disrotatory torsion of CH3ND3+, in kg·m^2."""
I_CD3ND3_dis = 1 / ((1/I_CD3) + (1/I_ND3))
"""Inertia of the disrotatory torsion of CD3ND3+, in kg·m^2."""

# Inertias, atomic units
I_CH3_amu = I_CH3 / (amu_to_kg * 1e-20)
"""Inertia of CH3, in amu·AA^2."""
I_CD3_amu = I_CD3 / (amu_to_kg * 1e-20)
"""Inertia of CD3, in amu·AA^2."""
I_NH3_amu = I_NH3 / (amu_to_kg * 1e-20)
"""Inertia of NH3, in amu·AA^2."""
I_ND3_amu = I_ND3 / (amu_to_kg * 1e-20)
"""Inertia of ND3, in amu·AA^2."""

# Inertias of the co-rotation, amu units
I_CH3NH3_amu = I_CH3_amu + I_NH3_amu
"""Inertia of CH3NH3+, in amu·AA^2."""
I_CD3NH3_amu = I_CD3_amu + I_NH3_amu
"""Inertia of CD3NH3+, in amu·AA^2."""
I_CH3ND3_amu = I_CH3_amu + I_ND3_amu
"""Inertia of CH3ND3+, in amu·AA^2."""
I_CD3ND3_amu = I_CD3_amu + I_ND3_amu
"""Inertia of CD3ND3+, in amu·AA^2."""

# Inertias of the dis-rotation, amu units
I_CH3NH3_dis_amu = I_CH3NH3_dis / (amu_to_kg * 1e-20)
"""Inertia of the disrotatory torsion of CH3NH3+, in amu·AA^2."""
I_CD3NH3_dis_amu = I_CD3NH3_dis / (amu_to_kg * 1e-20)
"""Inertia of the disrotatory torsion of CD3NH3+, in amu·AA^2."""
I_CH3ND3_dis_amu = I_CH3ND3_dis / (amu_to_kg * 1e-20)
"""Inertia of the disrotatory torsion of CH3ND3+, in amu·AA^2."""
I_CD3ND3_dis_amu = I_CD3ND3_dis / (amu_to_kg * 1e-20)
"""Inertia of the disrotatory torsion of CD3ND3+, in amu·AA^2."""

# Rotational energy
_hbar = const.physical_constants['reduced Planck constant'][0]
B_CH3 = ((_hbar**2) / (2 * I_CH3)) * (1000 / const.eV)
"""Kinetic rotational energy of CH3, in meV·s/kg·m^2."""
B_CD3 = ((_hbar**2) / (2 * I_CD3)) * (1000 / const.eV)
"""Kinetic rotational energy of CD3, in meV·s/kg·m^2."""
B_NH3 = ((_hbar**2) / (2 * I_NH3)) * (1000 / const.eV)
"""Kinetic rotational energy of NH3, in meV·s/kg·m^2."""
B_ND3 = ((_hbar**2) / (2 * I_ND3)) * (1000 / const.eV)
"""Kinetic rotational energy of ND3, in meV·s/kg·m^2."""

B_CH3NH3 = ((_hbar**2) / (2 * I_CH3NH3)) * (1000 / const.eV)
"""Kinetic rotational energy of CH3NH3+, in meV·s/kg·m^2."""
B_CD3NH3 = ((_hbar**2) / (2 * I_CD3NH3)) * (1000 / const.eV)
"""Kinetic rotational energy of CD3NH3+, in meV·s/kg·m^2."""
B_CH3ND3 = ((_hbar**2) / (2 * I_CH3ND3)) * (1000 / const.eV)
"""Kinetic rotational energy of CH3ND3+, in meV·s/kg·m^2."""
B_CD3ND3 = ((_hbar**2) / (2 * I_CD3ND3)) * (1000 / const.eV)
"""Kinetic rotational energy of CD3ND3+, in meV·s/kg·m^2."""

B_CH3NH3_dis = ((_hbar**2) / (2 * I_CH3NH3_dis)) * (1000 / const.eV)
"""Kinetic rotational energy of the disrotatory torsion of CH3NH3+, in meV·s/kg·m^2."""
B_CD3NH3_dis = ((_hbar**2) / (2 * I_CD3NH3_dis)) * (1000 / const.eV)
"""Kinetic rotational energy of the disrotatory torsion of CD3NH3+, in meV·s/kg·m^2."""
B_CH3ND3_dis = ((_hbar**2) / (2 * I_CH3ND3_dis)) * (1000 / const.eV)
"""Kinetic rotational energy of the disrotatory torsion of CH3ND3+, in meV·s/kg·m^2."""
B_CD3ND3_dis = ((_hbar**2) / (2 * I_CD3ND3_dis)) * (1000 / const.eV)
"""Kinetic rotational energy of the disrotatory torsion of CD3ND3+, in meV·s/kg·m^2."""

# Potential constants from titov2023 [C1, C2, C3, C4, C5]
constants_titov2023 = [
    [2.7860, 0.0130,-1.5284,-0.0037,-1.2791],  # ZIF-8
    [2.6507, 0.0158,-1.4111,-0.0007,-1.2547],  # ZIF-8 + Ar-1
    [2.1852, 0.0164,-1.0017, 0.0003,-1.2061],  # ZIF-8 + Ar-{1,2}
    [5.9109, 0.0258,-7.0152,-0.0168, 1.0213],  # ZIF-8 + Ar-{1,2,3}
    [1.4526, 0.0134,-0.3196, 0.0005,-1.1461],  # ZIF-8 + Ar-{1,2,4}
    ]
"""Potential constants from
[K. Titov et al., Phys. Rev. Mater. 7, 073402 (2023)](https://link.aps.org/doi/10.1103/PhysRevMaterials.7.073402)
for the `qrotor.potential.titov2023` potential.
In meV units.
"""




import qrotor as qr


def test_constants():
    assert round(qr.B_CH3, 5) == 0.64518
    assert round(qr.B_CD3, 5) == 0.32289
    assert round(qr.B_NH3, 5) == 0.73569
    assert round(qr.B_ND3, 5) == 0.36819
    assert round(qr.Ry_to_eV, 5)  == 13.60569
    assert round(qr.Ry_to_meV, 5) == 13605.69312
    assert round(qr.eV_to_Ry, 5)  == 0.07350
    assert round(qr.meV_to_Ry, 10) == .0000734986




import qrotor as qr


def test_tags():
    sys1 = qr.System(tags='tag1 tag2 tag3', comment='sys1', potential_name='zero')
    sys2 = qr.System(tags='tag2 tag3 tag4', comment='sys2')
    sys3 = qr.System(tags='tag4 tag5 tag6', comment='sys3')
    sys1.solve(100)
    test1 = qr.systems.filter_tags(sys1, include='tag4')
    assert test1 == []
    test2 = qr.systems.filter_tags([sys1, sys2], include='tag4 tag5', strict=False)
    assert len(test2) == 1
    assert test2[0].comment == 'sys2'
    test3 = qr.systems.filter_tags([sys1, sys2], include='tag4 tag5', strict=True)
    assert test3 == []
    test4 = qr.systems.filter_tags([sys1, sys2, sys3], include='tag3 tag4', strict=False)
    assert len(test4) == 3
    test5 = qr.systems.filter_tags([sys1, sys2, sys3], include='tag3 tag4', strict=True)
    assert len(test5) == 1
    assert test5[0].comment == 'sys2'
    test6 = qr.systems.filter_tags([sys1, sys2, sys3], include='tag3 tag4', exclude='tag6', strict=False)
    assert len(test6) == 2
    test7 = qr.systems.filter_tags([sys1, sys2, sys3], include='tag4', exclude='tag5 tag6', strict=True)
    assert len(test7) == 1
    assert test7[0].comment == 'sys2'
    test8 = qr.systems.filter_tags([sys1, sys2, sys3], include='', exclude='tag1 tag2', strict=True)
    assert len(test8) == 2
    test9 = qr.systems.filter_tags([sys1, sys2, sys3], include='', exclude='tag1 tag2', strict=False)
    assert test9[0].comment == 'sys3'

